{{
        Pseudo-DCT/Walsh Transform on the Propeller
        Jonathan "lonesock" Dummer

        The actual number of frequency buckets must be a multiple of 2.

        The distribute function takes ~6 PASM instructions per bucket.
        So at 80MHz clock, with 16kHz sample rate, you max out at ~200
        buckets.  Dropping to 8kHz or using a faster clock will buy you
        more, but probably isn't necessary, or desirable.

        Using a 6-point-cosine, the distribute function takes 13 PASM
        instructions per bucket.  So 80MHz / 16kHz / 13 = ~94.

        Frequency of a given bucket B:
        F = SampleRate*(B+1)/2/NumBuckets     
}}

CON
  ' sampling and frequency transform stuff
  NumBuckets    = 128
  SampleRate    = 8000
  ' target word storage and circular buffer stuff
  BufLen        = 4 * SampleRate / NumBuckets           ' 4 seconds worth of storage (at 8kHz/128, this is 1kB of storage)
  WordLen       = 32                                    ' All "words" will be resampled to this before comparison
  MaxWordSigs   = 10                                    ' How many words do I check against?
  ThresholdCount= SampleRate / NumBuckets / 4           ' 1/4 second
  ' index markers inside the transformed block
  idxLow        = (250  * NumBuckets * 2 / SampleRate) - 1
  idxMid        = (950  * NumBuckets * 2 / SampleRate) - 1
  idxHigh       = (2500 * NumBuckets * 2 / SampleRate) - 1
  ' Relative weights (log2) of the formants: F1,F2,F2/(F1+F2),Sib/(Sib+F1+F2)   

DAT  
  cbuf  long 0[BufLen]  ' storing audio data as it waits to be processed
  cog   byte 0          ' cog ID of the SPeechIN engine  

PUB start( adc_drive_pin, adc_monitor_pin, data_ptr ) | v
  stop
  ' set the data ptr to "not a word"
  long[data_ptr]~~
  ' set up my parameters before starting the cog
  cbuf_beg := @cbuf
  cbuf_ptr := cbuf_beg
  cbuf_end := cbuf_beg + (BufLen << 2)
  Nminus1 := NumBuckets - 1
  Nminus1times2 := Nminus1 << 1
  period := (clkfreq + (SampleRate >> 1)) / SampleRate
  dira_val := |<adc_drive_pin
  ' add in the headphones output?
  dira_val |= |<10 + |<11
  ' set up the ADC in counter A
  ctra_val := (%01001 << 26) | (adc_drive_pin << 9) | (adc_monitor_pin << 0)
  ' set up the borders for the formants
  split_points := idxLow               << 0
  split_points |= (idxMid-idxLow)      << 8
  split_points |= (idxHigh-idxMid)     << 16
  split_points |= (NumBuckets-idxHigh) << 24
  cog := cognew( @pasm_SPeechIN_entry, data_ptr ) + 1
  return cog

PUB stop
  if cog
    cogstop( cog~ - 1 ) 

{===== PASM Pseudo-Walsh Transform Engine =====}
DAT ORG 0
pasm_SPeechIN_entry
        ' get my ADC running
        mov dira,dira_val
        mov ctra,ctra_val
        mov frqa,#1
        ' and DAC too?
        movs ctrb,#10                       
        movd ctrb,#11
        movi ctrb,#%00111_000
        ' we are not yet recording
        mov recording,#0  
        ' clear out my accumulation "tank"
        call #clear_buckets

main_loop
        ' starting with the 1st sample
        mov sample_ID,#0
        ' set up timing, sampling, and grab a "last sample"
        mov timestamp,period
        add timestamp,cnt
        mov phsa,#0             
        waitcnt timestamp,period
        mov last_sample,phsa
        mov phsa,#0
        ' now grab as many samples as we need  
:grab_sample
        waitcnt timestamp,period
        mov sample,phsa
        mov phsa,#0
        ' use 1st derivative...removes DC, attenuates lower frequencies
        sub sample,last_sample
        add last_sample,sample
        ' output to headphones?
        mov tmp,sample
        shl tmp,#32-14
        add tmp,midscale
        mov frqb,tmp
        ' accumulate the data into the energy buckets
        call #distribute_sample
        ' move on to the next sample, and continue unless done
        add sample_ID,#1
        cmp sample_ID,Nminus1 wz,wc
if_be   jmp #:grab_sample
        ' do the processing
        call #process_data
        ' OJO - HACK
        mov wbuf,formants
        ' handle the state machine
        call #state_machine
        ' send out the debug info and clear out the buckets
        call #send_data_to_Hub
        call #clear_buckets 
        ' repeat
        jmp #main_loop


state_machine        
        ' are we above or below the floor level?
        cmp sum_all,floor_level wz,wc
if_ae   mov recording,#ThresholdCount

        ' we may be below the threshold, but waiting for a timeout
        cmpsub recording,#1 wc
if_nc   jmp #update_floor_level   

        ' we are above the floor level, so this is important, store it and advance the pointer
        wrlong formants,cbuf_ptr
        add cbuf_ptr,#4
        
        ' did we just run out of storage?
        cmp cbuf_ptr,cbuf_end wz,wc
if_ae   call #process_word
        jmp #state_machine_ret

update_floor_level
        ' we are below the floor, so adjust the floor level (move 1/N floor toward sum_all)
        neg tmp,floor_level
        sar tmp,#1                              ' 1 => 1/2 floor, 2 => 1/4 floor
        add tmp,sum_all wc
if_nc   sar tmp,#2 ' going down, adjust quickly
if_c    sar tmp,#8 ' going up, move slower
        add floor_level,tmp
        
        ' and did we just finish a word? I.e. we have data and it is long enough to make sense of?
        ' each formant is 4 bytes, so this is making sure we have at least WordLen/2 samples.
        mov tmp,cbuf_ptr
        sub tmp,cbuf_beg
        shr tmp,#2
        cmp tmp,#(WordLen/2+ThresholdCount) wz,wc
if_b    mov cbuf_ptr,cbuf_beg   ' reset my data pointer
if_b    mov recording,#0        ' reset my countdown          
if_ae   call #process_word                                                       
         
state_machine_ret
        ret


process_word
        ' resample the buffer into wbuf, of length WordLen
        movd :write_ws,#wbuf
        mov f_left,cbuf_ptr
        sub f_left,cbuf_beg
        shr f_left,#2
        sub f_left,#ThresholdCount+1
        ' reset my data pointer
        mov cbuf_ptr,cbuf_beg
        ' start out my accumulator
        mov accum,f_left
        shr accum,#1
        ' and how many slots are left to fill?
        mov tmp,#WordLen
:scale_step
        add accum,f_left
:baby_steps
        cmpsub accum,#WordLen-1 wc
if_c    add cbuf_ptr,#4
if_c    jmp #:baby_steps
        rdlong out_val,cbuf_ptr           
:write_ws
        mov 0-0,out_val
        add :write_ws,incDest
        djnz tmp,#:scale_step

        ' compare it to all stored samples

        ' reset my data pointer
        mov cbuf_ptr,cbuf_beg
        mov recording,#0
        
process_word_ret
        ret
        

''      SLOWER: 6-Point Cosine
'       (reduces spurious aliasing energy, compared to the 2-point cosine)     
'{        '' Inputs: sample, sample_ID
distribute_sample
        ' start with the first bucket
        movd :add_to_bucket,#buckets
        ' my accumulator is reset
        mov accum,Nminus1
        shr accum,#1
        ' how many buckets?
        mov f_left,#NumBuckets
        ' for my 6-point cosine, we add 3*sample_ID
        ' (start at 1, not 0...see below)
        mov six_point,#1
        mov step_add,sample_ID
        add step_add,sample_ID
        add step_add,sample_ID
        ' do the distribution
:this_bucket
        { NOTE:
          For the six point cosine we are using values:
          1  1/2  -1/2  -1  -1/2  1/2          
          which corresponds to cos(N*Pi/3).  However, to
          take advantage of mirroring the sign, I am using
          a rotated version (but starting at 1):
          1/2  1  1/2  -1/2  -1  -1/2  
          This way I can swap the sign every 3 steps, and
          use 1/2 if the index is 0 or 2.               }
        cmp six_point,#1 wz
        mov tmp,sample
if_nz   sar tmp,#1
:add_to_bucket
        adds 0-0,tmp
        add :add_to_bucket,incDest 
        ' do a single Bresenham step
        add accum,step_add
        ' (now a posssible 3 notches per step)
        cmpsub accum,Nminus1times2 wc
if_c    add six_point,#2
        cmpsub accum,Nminus1 wc
if_c    add six_point,#1
        ' update the sign of the sample is necessary
        cmpsub six_point,#3 wc
if_c    neg sample,sample
        ' go back for more
        djnz f_left,#:this_bucket
        ' done
distribute_sample_ret
        ret '}


''      FASTER: 2-Point Cosine
{        '' Inputs: sample, sample_ID
distribute_sample
        ' start with the first bucket
        movd :add_to_bucket,#buckets
        ' my accumulator is reset
        mov accum,Nminus1
        shr accum,#1
        ' how many buckets?
        mov tmp,#NumBuckets
        ' do the distribution
:add_to_bucket
        adds 0-0,sample
        add :add_to_bucket,incDest 
        ' do a single Bresenham step
        add accum,sample_ID
        cmpsub accum,Nminus1 wc
if_c    neg sample,sample
        djnz tmp,#:add_to_bucket
        ' done
distribute_sample_ret
        ret '}


process_data
        ' centroid of each back is stored as a pseudo-formant
        mov formants,#0
        mov sum_all,#0
        
        ' skip the first X buckets
        movs :get_v,#buckets        
        mov tmp,split_points
        and tmp,#$FF
        ror split_points,#8
        add :get_v,tmp
             
        ' do 2 centroid cals
        mov f_left,#2
:centroid
        mov sum_v,#0
        mov sum_iv,#0
        mov sample_ID,#0
        mov tmp,split_points
        and tmp,#$FF
        ror split_points,#8
        ' calculate the this centroid        
:get_v  abs op2,0-0
        add :get_v,#1           ' next bucket
        add sum_v,op2
        add sum_all,op2
        mov op1,sample_ID
        add sample_ID,#1
        call #peasant_multiply
        add sum_iv,out_val
        djnz tmp,#:get_v
        
        ' calculate the centroid value (pseudo-formant) using my
        ' the ratio method...multiply sum_v by the max lever-arm
        mov op1,sample_ID
        sub op1,#1
        mov op2,sum_v
        call #peasant_multiply
        ' now get the ratio of sum_iv / (sum_v*sample_ID)         
        mov op1,sum_iv
        mov op2,out_val
        call #compute_ratio_byte
        ' store this formant value        
        shl formants,#8
        or formants,out_val
        ' go back for more, if needed
        djnz f_left,#:centroid

        ' after the 2nd (last) centroid, store the ratio of the energy in F2 to the total so far
        mov op1,sum_v
        mov op2,sum_all
        call #compute_ratio_byte       
        shl formants,#8
        or formants,out_val
        
        ' OK, do the final summation for (hopefully) sibilance detection
        movs :add_last,:get_v
        mov sum_high,#0
        mov tmp,split_points
        and tmp,#$FF
        ror split_points,#8
:add_last
        addabs sum_high,0-0
        add :add_last,#1
        djnz tmp,#:add_last
        
        ' now get the ratio of sum_high / sum_all
        add sum_all,sum_high        
        mov op1,sum_high
        mov op2,sum_all
        call #compute_ratio_byte
        ' store this sibilance ratio value        
        shl formants,#8
        or formants,out_val
            
process_data_ret
        ret

        
send_data_to_Hub
        ' par holds thw write-to address
        mov sample,par 
        movd :send_bucket,#buckets
        mov tmp,#NumBuckets
:send_bucket
        wrlong 0-0,sample
        add :send_bucket,incDest
        add sample,#4
        djnz tmp,#:send_bucket

        ' send the formant long, replacing the 0th bucket
        wrlong formants,par
        'wrlong sum_all,par
        'wrlong floor_level,par 

        ' write out the last word signature
        movd :send_sig,#wbuf
        mov tmp,#WordLen
:send_sig
        wrlong 0-0,sample
        add :send_sig,incDest
        add sample,#4
        djnz tmp,#:send_sig
        
        ' done
send_data_to_Hub_ret
        ret       
        
clear_buckets 
        movd :zero_bucket,#buckets
        mov tmp,#NumBuckets
:zero_bucket
        mov 0-0,#0
        add :zero_bucket,incDest
        djnz tmp,#:zero_bucket
        ' done
clear_buckets_ret
        ret


{===== Peasant Multiplication =====}
' Stops on 0, works as long as the
' product fits in 32 bits. (So this
' is good for e.g. 24-bit x 5-bit.)
' All values treated as unsigned.
' 540 clocks max (including call & ret)
' for 8 bit op1's, 156 clocks
peasant_multiply
        ' put the smallest op in op1
        cmp op1,op2 wc,wz
if_b    xor op1,op2
if_b    xor op2,op1
if_b    xor op1,op2
        ' clear the accumulator
        mov out_val,#0
:loop   shr op2,#1 wc,wz
if_c    add out_val,op1
        shl op1,#1
if_nz   jmp #:loop
peasant_multiply_ret
        ret

        
' outval = 8 bit ratio of op1/op2
' 144 clocks (including call & ret)
compute_ratio_byte
        mov tmp,#8
        mov out_val,#0
:ratio_bit
        shl op1,#1
        cmpsub op1,op2 wc
        rcl out_val,#1
        djnz tmp,#:ratio_bit
compute_ratio_byte_ret
        ret
        

' outval := op1 / op2
' accum is the remainder
' 656 clocks (including call & ret)
{
long_divide
        mov accum,#0
        mov tmp,#32
:div_loop
        shl op1,#1 wc,wz
        rcl accum,#1
        cmpsub accum,op2 wc
        rcl out_val,#1
        djnz tmp,#:div_loop
long_divide_ret
        ret
'}

{===== PASM Parameters =====}
cbuf_ptr      long 0    ' where am I about to store this sample?
cbuf_beg      long 0    ' pointer to the start of the circular buffer
cbuf_end      long 0    ' pointer to the 1 past the end of the circular buffer
Nminus1       long 0    ' number of buckets - 1
Nminus1times2 long 0
dira_val      long 0    ' the output mask (for the drive pin on the ADC)
ctra_val      long 0    ' the entire counter control register to use the ADC counter mode
period        long 0    ' number of clock ticks between samples (clkfreq/Hz)
split_points  long 0    ' each byte holds the number of buckets for:
                        ' * low frequency
                        ' * F1 frequency range
                        ' * F2 frequency range
                        ' * the high frequency range (sibilance)

{===== PASM Constants and Initialized Variables =====}
floor_level    long |<30
incDest       long 512
'incDestAndSrc long 512 | 1

' for DAC
midscale      long |<31

{===== PASM Scratch Variables =====}
recording     res 1
six_point     res 1
step_add      res 1
timestamp     res 1
accum         res 1
tmp           res 1 
sample        res 1
last_sample   res 1
sample_ID     res 1
f_left        res 1
sum_v         res 1
sum_iv        res 1
op1           res 1
op2           res 1
out_val       res 1
sum_all       res 1
sum_high      res 1

{===== Accumulation Buckets for each frequency =====}
buckets       res NumBuckets    ' storing the transformed data
formants      res 1             ' store the 4 values computed per block (1 byte each)
wbuf          res WordLen       ' storing the resampled processed data, waiting for comparison

FIT 496